import pathlib
from math import fsum

import numpy as np
import pytest

import meshplex

from .helpers import is_near_equal, run

this_dir = pathlib.Path(__file__).resolve().parent


@pytest.mark.parametrize("a", [0.5, 1.0, 1.33])  # edge length
def test_regular_tet0(a):
    points = (
        a
        * np.array(
            [
                [1.0, 0, 0],
                [-0.5, +np.sqrt(3.0) / 2.0, 0],
                [-0.5, -np.sqrt(3.0) / 2.0, 0],
                [0.0, 0.0, np.sqrt(2.0)],
            ]
        )
        / np.sqrt(3.0)
    )
    cells = np.array([[0, 1, 2, 3]])
    mesh = meshplex.MeshTetra(points, cells.copy())
    # test __repr__
    print(mesh)

    assert np.all(mesh.cells("points") == cells)

    mesh.show()
    mesh.show_edge(0)

    tol = 1.0e-14

    z = a / np.sqrt(24.0)
    assert is_near_equal(mesh.cell_circumcenters, [0.0, 0.0, z], tol)

    assert is_near_equal(mesh.circumcenter_facet_distances, [z, z, z, z], tol)

    # covolume/edge length ratios
    # alpha = a / 12.0 / np.sqrt(2)
    alpha = a / 2 / np.sqrt(24) / np.sqrt(12)
    vals = mesh.ce_ratios
    assert is_near_equal(
        vals,
        [
            [
                [alpha, alpha, alpha],
                [alpha, alpha, alpha],
                [alpha, alpha, alpha],
                [alpha, alpha, alpha],
            ]
        ],
        tol,
    )

    # cell volumes
    vol = a**3 / 6.0 / np.sqrt(2)
    assert is_near_equal(mesh.cell_volumes, [vol], tol)

    # control volumes
    val = vol / 4.0
    assert is_near_equal(mesh.control_volumes, [val, val, val, val], tol)

    # inradius
    # side_area = np.sqrt(3) / 4 * a ** 2
    # vol = a ** 3 / 6.0 / np.sqrt(2)
    # inradius = 3 * vol / (4 * side_area)
    inradius = a * np.sqrt(6) / 12
    assert is_near_equal(mesh.cell_inradius, [inradius], tol)

    # circumradius
    circumradius = a * np.sqrt(6) / 4
    assert is_near_equal(mesh.cell_circumradius, [circumradius], tol)

    # cell quality
    assert is_near_equal(mesh.q_radius_ratio, [1.0], tol)

    assert is_near_equal(mesh.cell_barycenters, mesh.cell_centroids, tol)
    assert is_near_equal(mesh.cell_barycenters, [[0.0, 0.0, a * np.sqrt(6) / 12]], tol)

    assert is_near_equal(mesh.cell_incenters, [[0.0, 0.0, a * np.sqrt(6) / 12]], tol)

    assert is_near_equal(mesh.q_min_sin_dihedral_angles, [1.0], tol)

    assert is_near_equal(mesh.q_vol_rms_edgelength3, [1.0], tol)


# @pytest.mark.parametrize(
#         'a',  # basis edge length
#         [1.0]
#         )
# def test_regular_tet1_algebraic(a):
#     points = np.array([
#         [0, 0, 0],
#         [a, 0, 0],
#         [0, a, 0],
#         [0, 0, a]
#         ])
#     cells = np.array([[0, 1, 2, 3]])
#     tol = 1.0e-14
#
#     mesh = meshplex.MeshTetra(points, cells, mode='algebraic')
#
#     assert is_near_equal(
#         mesh.cell_circumcenters,
#         [[a/2.0, a/2.0, a/2.0]],
#         tol
#         )
#
#     # covolume/edge length ratios
#     assert is_near_equal(
#         mesh.ce_ratios_per_edge,
#         [a/6.0, a/6.0, a/6.0, 0.0, 0.0, 0.0],
#         tol
#         )
#
#     # cell volumes
#     assert is_near_equal(mesh.cell_volumes, [a**3/6.0], tol)
#
#     # control volumes
#     assert is_near_equal(
#         mesh.control_volumes,
#         [a**3/12.0, a**3/36.0, a**3/36.0, a**3/36.0],
#         tol
#         )


@pytest.mark.parametrize(
    "a",
    [
        # 0.5,
        1.0,
        # 2.0
    ],
)  # basis edge length
def test_unit_tetrahedron_geometric(a):
    points = np.array([[0, 0, 0], [a, 0, 0], [0, a, 0], [0, 0, a]])
    cells = np.array([[0, 1, 2, 3]])
    tol = 1.0e-14

    mesh = meshplex.MeshTetra(points, cells)

    assert all((mesh.cells("points") == cells).flat)

    assert is_near_equal(mesh.cell_circumcenters, [a / 2.0, a / 2.0, a / 2.0], tol)

    # covolume/edge length ratios
    ref = np.array(
        [
            [[-a / 24.0], [a / 8.0], [a / 8.0], [0.0]],
            [[-a / 24.0], [a / 8.0], [0.0], [a / 8.0]],
            [[-a / 24.0], [0.0], [a / 8.0], [a / 8.0]],
        ]
    )
    assert is_near_equal(mesh.ce_ratios, ref, tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [a**3 / 6.0], tol)

    # control volumes
    assert is_near_equal(
        mesh.control_volumes,
        [a**3 / 8.0, a**3 / 72.0, a**3 / 72.0, a**3 / 72.0],
        tol,
    )

    assert is_near_equal(
        mesh.circumcenter_facet_distances.T,
        [-0.5 / np.sqrt(3) * a, 0.5 * a, 0.5 * a, 0.5 * a],
        tol,
    )

    # inradius
    ref = a * 0.2113248654051872
    assert is_near_equal(mesh.cell_inradius, [ref], tol)

    # circumradius
    ref = a * np.sqrt(3) / 2
    assert is_near_equal(mesh.cell_circumradius, [ref], tol)

    # cell quality
    ref = 7.320508075688774e-01
    assert is_near_equal(mesh.q_radius_ratio, [ref], tol)

    assert is_near_equal(mesh.q_min_sin_dihedral_angles, [np.sqrt(3) / 2], tol)

    assert is_near_equal(mesh.q_vol_rms_edgelength3, [4 * np.sqrt(3) / 9], tol)


def test_regular_tet1_geometric_order():
    a = 1.0
    points = np.array([[0, 0, a], [0, 0, 0], [a, 0, 0], [0, a, 0]])
    cells = np.array([[0, 1, 2, 3]])
    tol = 1.0e-14

    mesh = meshplex.MeshTetra(points, cells)

    assert all((mesh.cells("points") == [0, 1, 2, 3]).flat)

    assert is_near_equal(mesh.cell_circumcenters, [a / 2.0, a / 2.0, a / 2.0], tol)

    # covolume/edge length ratios
    ref = np.array(
        [
            [[0.0], [-a / 24.0], [a / 8.0], [a / 8.0]],
            [[a / 8.0], [-a / 24.0], [a / 8.0], [0.0]],
            [[a / 8.0], [-a / 24.0], [0.0], [a / 8.0]],
        ]
    )
    assert is_near_equal(mesh.ce_ratios, ref, tol)

    # cell volumes
    assert is_near_equal(mesh.cell_volumes, [a**3 / 6.0], tol)

    # control volumes
    assert is_near_equal(
        mesh.control_volumes,
        [a**3 / 72.0, a**3 / 8.0, a**3 / 72.0, a**3 / 72.0],
        tol,
    )

    assert is_near_equal(
        mesh.circumcenter_facet_distances.T,
        [0.5 * a, -0.5 / np.sqrt(3) * a, 0.5 * a, 0.5 * a],
        tol,
    )


# @pytest.mark.parametrize(
#         'h',
#         [1.0, 1.0e-2]
#         )
# def test_degenerate_tet0(h):
#     points = np.array([
#         [0, 0, 0],
#         [1, 0, 0],
#         [0, 1, 0],
#         [0.5, 0.5, h],
#         ])
#     cells = np.array([[0, 1, 2, 3]])
#     mesh = meshplex.MeshTetra(points, cells, mode='algebraic')
#
#     tol = 1.0e-14
#
#     z = 0.5 * h - 1.0 / (4*h)
#     assert is_near_equal(
#         mesh.cell_circumcenters,
#         [[0.5, 0.5, z]],
#         tol
#         )
#
#     # covolume/edge length ratios
#     print(h)
#     print(mesh.ce_ratios)
#     assert is_near_equal(
#         mesh.ce_ratios,
#         [[
#             [0.0, 0.0, 0.0],
#             [3.0/80.0, 3.0/40.0, 3.0/80.0],
#             [3.0/40.0, 3.0/80.0, 3.0/80.0],
#             [0.0, 1.0/16.0, 1.0/16.0],
#         ]],
#         tol
#         )
#     # [h / 6.0, h / 6.0, 0.0, -1.0/24/h, 1.0/12/h, 1.0/12/h],
#
#     # control volumes
#     ref = [
#         h / 18.0,
#         1.0/72.0 * (3*h - 1.0/(2*h)),
#         1.0/72.0 * (3*h - 1.0/(2*h)),
#         1.0/36.0 * (h + 1.0/(2*h))
#         ]
#     assert is_near_equal(mesh.control_volumes, ref, tol)
#
#     # cell volumes
#     assert is_near_equal(mesh.cell_volumes, [h/6.0], tol)


# @pytest.mark.parametrize(
#         'h',
#         [1.0e-1]
#         )
# def test_degenerate_tet1(h):
#     points = np.array([
#         [0, 0, 0],
#         [1, 0, 0],
#         [0, 1, 0],
#         [0.25, 0.25, h],
#         [0.25, 0.25, -h],
#         ])
#     cells = np.array([
#         [0, 1, 2, 3],
#         [0, 1, 2, 4]
#         ])
#     mesh = meshplex.MeshTetra(points, cells, mode='algebraic')
#
#     total_vol = h / 3.0
#
#     run(
#         mesh,
#         total_vol,
#         [0.18734818957173291, 77.0/720.0],
#         [2.420625, 5.0/6.0],
#         [1.0 / np.sqrt(2.0) / 30., 1.0/60.0]
#         )


def test_cubesmall():
    points = np.array(
        [
            [-0.5, -0.5, -5.0],
            [-0.5, +0.5, -5.0],
            [+0.5, -0.5, -5.0],
            [-0.5, -0.5, +5.0],
            [+0.5, +0.5, -5.0],
            [+0.5, +0.5, +5.0],
            [-0.5, +0.5, +5.0],
            [+0.5, -0.5, +5.0],
        ]
    )
    cells = np.array(
        [[0, 1, 2, 3], [1, 2, 4, 5], [1, 2, 3, 5], [1, 3, 5, 6], [2, 3, 5, 7]]
    )
    mesh = meshplex.MeshTetra(points, cells)

    tol = 1.0e-14

    cv = np.ones(8) * 1.25
    cellvols = [5.0 / 3.0, 5.0 / 3.0, 10.0 / 3.0, 5.0 / 3.0, 5.0 / 3.0]

    assert is_near_equal(mesh.control_volumes, cv, tol)
    assert is_near_equal(mesh.cell_volumes, cellvols, tol)

    cv_norms = [np.linalg.norm(cv, ord=2), np.linalg.norm(cv, ord=np.Inf)]
    cellvol_norms = [
        np.linalg.norm(cellvols, ord=2),
        np.linalg.norm(cellvols, ord=np.Inf),
    ]
    run(mesh, 10.0, cv_norms, [28.095851618771825, 1.25], cellvol_norms)


def test_arrow3d():
    points = np.array(
        [
            [+0.0, +0.0, +0.0],
            [+2.0, -1.0, +0.0],
            [+2.0, +0.0, +0.0],
            [+2.0, +1.0, +0.0],
            [+0.5, +0.0, -0.9],
            [+0.5, +0.0, +0.9],
        ]
    )
    cellsNodes = np.array([[1, 2, 4, 5], [2, 3, 4, 5], [0, 1, 4, 5], [0, 3, 4, 5]])
    mesh = meshplex.MeshTetra(points, cellsNodes)

    run(
        mesh,
        1.2,
        [0.58276428453480922, 0.459],
        [0.40826901831985885, 0.2295],
        [np.sqrt(0.45), 0.45],
    )

    assert mesh.num_delaunay_violations == 2


def test_tetrahedron():
    mesh = meshplex.read(this_dir / "meshes" / "tetrahedron.vtk")

    run(
        mesh,
        64.1500299099584,
        [16.308991595922095, 7.0264329635751395],
        [6.898476155562041, 0.34400453539215237],
        [11.571692332290635, 2.9699087921277054],
    )


# def test_toy_algebraic():
#     filename = download_mesh(
#         'toy.vtk',
#         'f48abda972822bab224b91a74d695573'
#         )
#     mesh, _, _, _ = meshplex.read(filename)
#
#     # Even if the input data has only a small error, the error in the
#     # ce_ratios can be magnitudes larger. This is demonstrated here: Take
#     # the same mesh from two different source files with a difference of
#     # the order of 1e-16. The ce_ratios differ by up to 1e-7.
#     if False:
#         print(mesh.cells.keys())
#         pts = mesh.points.copy()
#         pts += 1.0e-16 * np.random.rand(pts.shape[0], pts.shape[1])
#         mesh2 = meshplex.MeshTetra(pts, mesh.cells['points'])
#         #
#         diff_coords = mesh.points - mesh2.points
#         max_diff_coords = max(diff_coords.flatten())
#         print('||coords_1 - coords_2||_inf  =  %e' % max_diff_coords)
#         diff_ce_ratios = mesh.ce_ratios_per_edge - mesh2.ce_ratios
#         print(
#             '||ce_ratios_1 - ce_ratios_2||_inf  =  %e'
#             % max(diff_ce_ratios)
#             )
#         from matplotlib import pyplot as plt
#         plt.figure()
#         n = len(mesh.ce_ratios_per_edge)
#         plt.semilogy(range(n), diff_ce_ratios)
#         plt.show()
#         exit(1)
#
#     run(
#         mesh,
#         volume=9.3875504672601107,
#         convol_norms=[0.20348466631551548, 0.010271101930468585],
#         ce_ratio_norms=[396.4116343366758, 3.4508458933423918],
#         cellvol_norms=[0.091903119589148916, 0.0019959463063558944],
#         tol=1.0e-6
#         )


def test_toy_geometric():
    mesh = meshplex.read(this_dir / "meshes" / "toy.vtu")

    mesh = meshplex.MeshTetra(mesh.points, mesh.cells("points"))

    run(
        mesh,
        volume=1.2558406318969064,
        convol_norms=[0.06426267704764405, 0.006376495345775972],
        ce_ratio_norms=[2.3838036244886744, 0.061790574680317234],
        cellvol_norms=[0.03813510075830597, 0.002568756890638246],
        tol=1.0e-6,
    )

    cc = mesh.cell_circumcenters
    cc_norm_2 = fsum(cc.flat)
    cc_norm_inf = max(cc.flat)
    assert abs(cc_norm_2 - 250.71142559136365) < 1.0e-11, cc_norm_2
    assert abs(cc_norm_inf - 1.6682679741134048) < 1.0e-11, cc_norm_inf


def test_show_cell(render=False):
    pytest.importorskip("vtk")

    points = np.array(
        [
            [1.0, 0.0, -1.0 / np.sqrt(8)],
            [-0.5, +np.sqrt(3.0) / 2.0, -1.0 / np.sqrt(8)],
            [-0.5, -np.sqrt(3.0) / 2.0, -1.0 / np.sqrt(8)],
            [0.0, 0.0, np.sqrt(2.0) - 1.0 / np.sqrt(8)],
        ]
    ) / np.sqrt(3.0)

    # points = np.array(
    #     [
    #         [1.0, 0.0, -1.0 / np.sqrt(8)],
    #         [-0.5, +np.sqrt(3.0) / 2.0, -1.0 / np.sqrt(8)],
    #         [-0.5, -np.sqrt(3.0) / 2.0, -1.0 / np.sqrt(8)],
    #         [0.0, 0.5, 1.0],
    #     ]
    # ) / np.sqrt(3.0)
    # points = np.array(
    #     [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    # )

    cells = [[0, 1, 2, 3]]
    mesh = meshplex.MeshTetra(points, cells)
    mesh.show_cell(
        0,
        barycenter_rgba=(1, 0, 0, 1.0),
        circumcenter_rgba=(0.1, 0.1, 0.1, 1.0),
        circumsphere_rgba=(0, 1, 0, 1.0),
        incenter_rgba=(1, 0, 1, 1.0),
        insphere_rgba=(1, 0, 1, 1.0),
        face_circumcenter_rgba=(0, 0, 1, 1.0),
        control_volume_boundaries_rgba=(1.0, 0.0, 0.0, 1.0),
        line_width=3.0,
        close=True,
        render=render,
    )


if __name__ == "__main__":
    test_show_cell(render=True)
    # test_regular_tet0(0.5)
    # test_toy_geometric()



import tempfile

import meshzoo
import numpy as np

import meshplex


def test_io_2d():
    vertices, cells = meshzoo.rectangle_tri(
        np.linspace(0.0, 1.0, 3), np.linspace(0.0, 1.0, 3)
    )
    mesh = meshplex.MeshTri(vertices, cells)
    # mesh = meshplex.read('pacman.vtu')
    assert mesh.num_delaunay_violations == 0

    # mesh.show(show_axes=False, boundary_edge_color="g")
    # mesh.show_vertex(0)

    with tempfile.TemporaryDirectory() as tmpdir:
        mesh.write(tmpdir + "test.vtk")
        mesh2 = meshplex.read(tmpdir + "test.vtk")

    assert np.all(mesh.cells("points") == mesh2.cells("points"))


# def test_io_3d(self):
#     vertices, cells = meshzoo.cube(
#             0.0, 1.0, 0.0, 1.0, 0.0, 1.0,
#             2, 2, 2
#             )
#     mesh = meshplex.MeshTetra(vertices, cells)

#     self.assertEqual(mesh.num_delaunay_violations, 0)

#     # mesh.show_control_volume(0)
#     # mesh.show_edge(0)
#     # import matplotlib.pyplot as plt
#     # plt.show()

#     mesh.write('test.vtu')

#     mesh2, _, _, _ = meshplex.read('test.vtu')

#     for k in range(len(mesh.cells['points'])):
#         self.assertEqual(
#                 tuple(mesh.cells['points'][k]),
#                 tuple(mesh2.cells['points'][k])
#                 )



import numpy as np
import pytest

import meshplex


@pytest.mark.skip()
def test_line():
    X = np.array([0.0, 0.1, 0.4, 1.0])
    cells = np.array([[0, 1], [1, 2], [2, 3]])
    mesh = meshplex.Mesh(X, cells)
    mesh.create_facets()

    ref = np.array([0, 1, 2, 3])
    assert np.all(mesh.facets["points"] == ref)

    ref = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
        ]
    )
    assert np.all(mesh.cells("facets") == ref)


def test_tri():
    X = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2], [0, 2, 3]])
    mesh = meshplex.Mesh(X, cells)
    mesh.create_facets()

    ref = np.array(
        [
            [0, 1],
            [0, 2],
            [0, 3],
            [1, 2],
            [2, 3],
        ]
    )
    assert np.all(mesh.facets["points"] == ref)

    ref = np.array(
        [
            [3, 1, 0],
            [4, 2, 1],
        ]
    )
    assert np.all(mesh.cells("facets") == ref)


def test_tetra():
    X = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, -1.0],
            [0.0, 0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2, 3], [0, 1, 2, 4]])
    mesh = meshplex.Mesh(X, cells)
    mesh.create_facets()

    ref = np.array(
        [
            [0, 1, 2],
            [0, 1, 3],
            [0, 1, 4],
            [0, 2, 3],
            [0, 2, 4],
            [1, 2, 3],
            [1, 2, 4],
        ]
    )
    assert np.all(mesh.facets["points"] == ref)

    ref = np.array(
        [
            [5, 3, 1, 0],
            [6, 4, 2, 0],
        ]
    )
    assert np.all(mesh.cells("facets") == ref)


def test_duplicate_cells():
    X = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ]
    )
    cells = np.array([[0, 1, 2], [0, 2, 3], [0, 2, 1]])
    mesh = meshplex.Mesh(X, cells)
    with pytest.raises(meshplex.MeshplexError):
        mesh.create_facets()



import numpy as np

import meshplex


def assert_mesh_consistency(mesh0, tol=1.0e-14):
    assert np.all(np.logical_xor(mesh0.is_boundary_facet, mesh0.is_interior_facet))

    bpts = np.array(
        [
            mesh0.is_boundary_point,
            mesh0.is_interior_point,
            ~mesh0.is_point_used,
        ]
    )
    assert np.all(np.sum(bpts, axis=0) == 1)

    # consistency check for facets_cells
    assert np.all(mesh0.is_boundary_facet[mesh0.facets_cells["boundary"][0]])
    assert not np.any(mesh0.is_boundary_facet[mesh0.facets_cells["interior"][0]])

    for edge_id, cell_id, local_edge_id in mesh0.facets_cells["boundary"].T:
        assert edge_id == mesh0.cells("facets")[cell_id][local_edge_id]

    for (
        edge_id,
        cell_id0,
        cell_id1,
        local_edge_id0,
        local_edge_id1,
    ) in mesh0.facets_cells["interior"].T:
        assert edge_id == mesh0.cells("facets")[cell_id0][local_edge_id0]
        assert edge_id == mesh0.cells("facets")[cell_id1][local_edge_id1]

    # check consistency of facets_cells_idx with facets_cells
    for edge_id, idx in enumerate(mesh0.facets_cells_idx):
        if mesh0.is_boundary_facet[edge_id]:
            assert edge_id == mesh0.facets_cells["boundary"][0, idx]
        else:
            assert edge_id == mesh0.facets_cells["interior"][0, idx]

    # Assert facets_cells integrity
    for cell_gid, edge_gids in enumerate(mesh0.cells("facets")):
        for edge_gid in edge_gids:
            idx = mesh0.facets_cells_idx[edge_gid]
            if mesh0.is_boundary_facet[edge_gid]:
                assert cell_gid == mesh0.facets_cells["boundary"][1, idx]
            else:
                assert cell_gid in mesh0.facets_cells["interior"][1:3, idx]

    # make sure the edges are opposite of the points
    for cell_gid, (point_ids, edge_ids) in enumerate(
        zip(mesh0.cells("points"), mesh0.cells("facets"))
    ):
        for k in range(len(point_ids)):
            assert set(point_ids) == {*mesh0.edges["points"][edge_ids][k], point_ids[k]}

    # make sure the is_boundary_point/edge/cell is consistent
    ref_cells = np.any(mesh0.is_boundary_facet_local, axis=0)
    assert np.all(mesh0.is_boundary_cell == ref_cells)
    ref_points = np.zeros(len(mesh0.points), dtype=bool)
    ref_points[mesh0.idx[1][:, mesh0.is_boundary_facet_local]] = True
    assert np.all(mesh0.is_boundary_point == ref_points)

    assert len(mesh0.control_volumes) == len(mesh0.points)
    assert len(mesh0.control_volume_centroids) == len(mesh0.points)

    # TODO add more consistency checks

    # now check the numerical values
    # create a new mesh from the points and cells and compare
    mesh1 = meshplex.Mesh(mesh0.points, mesh0.cells("points"))

    # Can't add those tests since the facet order will be different.
    # TODO bring back
    # if mesh0.facets is None:
    #     mesh0.create_facets()
    # if mesh1.facets is None:
    #     mesh1.create_facets()
    # assert np.all(mesh0.boundary_facets == mesh1.boundary_facets)
    # assert np.all(mesh0.interior_facets == mesh1.interior_facets)
    # assert np.all(mesh0.is_boundary_facet == mesh1.is_boundary_facet)
    # assert np.all(
    #     np.abs(
    #         mesh0.signed_circumcenter_distances - mesh1.signed_circumcenter_distances
    #     )
    #     < tol
    # )

    assert np.all(mesh0.is_point_used == mesh1.is_point_used)
    assert np.all(mesh0.is_boundary_point == mesh1.is_boundary_point)
    assert np.all(mesh0.is_interior_point == mesh1.is_interior_point)
    assert np.all(mesh0.is_boundary_facet_local == mesh1.is_boundary_facet_local)
    assert np.all(mesh0.is_boundary_cell == mesh1.is_boundary_cell)

    assert np.all(np.abs(mesh0.ei_dot_ei - mesh1.ei_dot_ei) < tol)
    assert np.all(np.abs(mesh0.cell_volumes - mesh1.cell_volumes) < tol)
    assert np.all(
        np.abs(mesh0.circumcenter_facet_distances - mesh1.circumcenter_facet_distances)
        < tol
    )

    assert np.all(np.abs(mesh0.signed_cell_volumes - mesh1.signed_cell_volumes) < tol)
    assert np.all(np.abs(mesh0.cell_centroids - mesh1.cell_centroids) < tol)
    assert np.all(np.abs(mesh0.cell_circumcenters - mesh1.cell_circumcenters) < tol)
    assert np.all(np.abs(mesh0.control_volumes - mesh1.control_volumes) < tol)
    assert np.all(np.abs(mesh0.ce_ratios - mesh1.ce_ratios) < tol)

    ipu = mesh0.is_point_used
    assert np.all(
        np.abs(
            mesh0.control_volume_centroids[ipu] - mesh1.control_volume_centroids[ipu]
        )
        < tol
    )


def compute_all_entities(mesh):
    mesh.is_boundary_point
    mesh.is_interior_point
    mesh.is_boundary_facet_local
    mesh.is_boundary_facet
    mesh.is_boundary_cell
    mesh.cell_volumes
    mesh.ce_ratios
    mesh.signed_cell_volumes
    mesh.cell_centroids
    mesh.control_volumes
    mesh.create_facets()
    mesh.facets_cells
    mesh.facets_cells_idx
    mesh.boundary_facets
    mesh.interior_facets
    mesh.cell_circumcenters
    mesh.signed_circumcenter_distances
    mesh.control_volume_centroids

    assert mesh.edges is not None
    assert mesh.subdomains is not {}
    assert mesh._is_interior_point is not None
    assert mesh._is_boundary_point is not None
    assert mesh._is_boundary_facet_local is not None
    assert mesh._is_boundary_facet is not None
    assert mesh._is_boundary_cell is not None
    assert mesh._facets_cells is not None
    assert mesh._facets_cells_idx is not None
    assert mesh._boundary_facets is not None
    assert mesh._interior_facets is not None
    assert mesh._is_point_used is not None
    assert mesh._half_edge_coords is not None
    assert mesh._ei_dot_ei is not None
    assert mesh._volumes is not None
    assert mesh._ce_ratios is not None
    assert mesh._circumcenters is not None
    assert mesh._circumcenter_facet_distances is not None
    assert mesh._signed_circumcenter_distances is not None
    assert mesh._control_volumes is not None
    assert mesh._cell_partitions is not None
    assert mesh._cv_centroids is not None
    assert mesh._signed_cell_volumes is not None
    assert mesh._cell_centroids is not None



"""
"""
import random

import numpy as np
import perfplot
from scipy.spatial import Delaunay

import meshplex


def setup(n):
    radius = 1.0
    k = np.arange(n)
    boundary_pts = radius * np.column_stack(
        [np.cos(2 * np.pi * k / n), np.sin(2 * np.pi * k / n)]
    )

    # Compute the number of interior points such that all triangles can be somewhat
    # equilateral.
    edge_length = 2 * np.pi * radius / n
    domain_area = np.pi - n * (radius**2 / 2 * (edge_length - np.sin(edge_length)))
    cell_area = np.sqrt(3) / 4 * edge_length**2
    target_num_cells = domain_area / cell_area
    # Euler:
    # 2 * num_points - num_boundary_edges - 2 = num_cells
    # <=>
    # num_interior_points ~= 0.5 * (num_cells + num_boundary_edges) + 1 - num_boundary_points
    m = int(0.5 * (target_num_cells + n) + 1 - n)

    # generate random points in circle;
    # <https://mathworld.wolfram.com/DiskPointPicking.html>
    for seed in range(0, 255):
        np.random.seed(seed)
        r = np.random.rand(m)
        alpha = 2 * np.pi * np.random.rand(m)

        interior_pts = np.column_stack(
            [np.sqrt(r) * np.cos(alpha), np.sqrt(r) * np.sin(alpha)]
        )

        pts = np.concatenate([boundary_pts, interior_pts])

        tri = Delaunay(pts)

        # Make sure there are exactly `n` boundary points
        mesh0 = meshplex.MeshTri(pts, tri.simplices)
        mesh1 = meshplex.MeshTri(pts, tri.simplices)
        if np.sum(mesh0.is_boundary_point) == n:
            break

    mesh0.create_edges()
    mesh1.create_edges()

    num_interior_edges = np.sum(mesh0.is_interior_edge)
    idx = random.sample(range(num_interior_edges), n // 10)
    print(num_interior_edges, len(idx), len(idx) / num_interior_edges)

    # # move interior points a little bit such that we have edges to flip
    # max_step = np.min(mesh.cell_inradius) / 2
    # mesh.points = mesh.points + max_step * np.random.rand(*mesh.points.shape)
    # print(mesh.num_delaunay_violations)
    return mesh0, mesh1, idx


def flip_old(data):
    mesh0, mesh1, idx = data
    mesh0.flip_interior_edges_old(idx)


def flip_new(data):
    mesh0, mesh1, idx = data
    mesh1.flip_interior_edges(idx)


perfplot.show(
    setup=setup,
    kernels=[flip_old, flip_new],
    n_range=[2**k for k in range(5, 13)],
    equality_check=None,
    # set target time to 0 to avoid more than one repetition
    target_time_per_measurement=0.0,
)



from .__about__ import __version__
from ._exceptions import MeshplexError
from ._mesh import Mesh
from ._mesh_tetra import MeshTetra
from ._mesh_tri import MeshTri
from ._reader import from_meshio, read

__all__ = [
    "Mesh",
    "MeshTri",
    "MeshTetra",
    "MeshplexError",
    "read",
    "from_meshio",
    "__version__",
]



import pathlib

import meshplex

this_dir = pathlib.Path(__file__).resolve().parent


def test_show_mesh():
    mesh = meshplex.read(this_dir / ".." / "meshes" / "pacman.vtu")
    mesh = meshplex.MeshTri(mesh.points[:, :2], mesh.cells("points"))
    print(mesh)  # test __repr__
    # mesh.plot(show_axes=False)
    mesh.show(
        show_axes=False,
        cell_quality_coloring=("viridis", 0.0, 1.0, True),
        show_point_numbers=True,
        show_edge_numbers=True,
        show_cell_numbers=True,
        mark_points=[1],
        mark_edges=[0],
        mark_cells=[0, 3, 7],
        nondelaunay_edge_color="r",
        boundary_edge_color="b",
        control_volume_centroid_color="g",
    )
    # mesh.save("pacman.png", show_axes=False)


def test_show_vertex():
    mesh = meshplex.read(this_dir / ".." / "meshes" / "pacman.vtu")
    # mesh.plot_vertex(125)
    mesh.show_vertex(125)


if __name__ == "__main__":
    test_show_mesh()



import pathlib

import meshio
import numpy as np
import pytest

import meshplex

from .helpers import assert_mesh_consistency, compute_all_entities

this_dir = pathlib.Path(__file__).resolve().parent


def test_flip_simple():
    #        3                   3
    #        A                   A
    #       /|\                 / \
    #     1/ | \4             1/ 1 \4
    #     /  |  \             /     \
    #   0/ 0 3   \2   ==>   0/___3___\2
    #    \   | 1 /           \       /
    #     \  |  /             \     /
    #     0\ | /2             0\ 0 /2
    #       \|/                 \ /
    #        V                   V
    #        1                   1
    #
    points = np.array([[-0.1, 0.0], [0.0, -1.0], [0.1, 0.0], [0.0, 1.1]])
    cells = np.array([[0, 1, 3], [1, 2, 3]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.create_facets()
    assert not mesh.is_delaunay
    assert mesh.num_delaunay_violations == 1
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    assert np.array_equal(mesh.cells("edges"), [[3, 1, 0], [4, 3, 2]])
    assert_mesh_consistency(mesh)

    # mesh.show()
    num_flips = mesh.flip_until_delaunay()
    assert num_flips == 1

    assert_mesh_consistency(mesh)
    assert mesh.num_delaunay_violations == 0
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [0, 2], [2, 3]]
    )
    assert np.array_equal(mesh.cells("points"), [[0, 1, 2], [0, 2, 3]])
    assert np.array_equal(mesh.cells("edges"), [[2, 3, 0], [4, 1, 3]])


def test_flip_simple_negative_orientation():
    #        3                   3
    #        A                   A
    #       /|\                 / \
    #     1/ | \4             1/ 1 \4
    #     /  |  \             /     \
    #   0/ 0 3   \2   ==>   0/___3___\2
    #    \   | 1 /           \       /
    #     \  |  /             \     /
    #     0\ | /2             0\ 0 /2
    #       \|/                 \ /
    #        V                   V
    #        1                   1
    #
    points = np.array([[-0.1, 0.0], [0.0, -1.0], [0.1, 0.0], [0.0, 1.1]])
    cells = np.array([[0, 3, 1], [1, 3, 2]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.create_facets()
    assert mesh.num_delaunay_violations == 1
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    assert np.array_equal(mesh.cells("edges"), [[3, 0, 1], [4, 2, 3]])
    assert_mesh_consistency(mesh)

    # mesh.show()
    mesh.flip_until_delaunay()
    assert_mesh_consistency(mesh)
    assert mesh.num_delaunay_violations == 0
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [0, 2], [2, 3]]
    )
    assert np.array_equal(mesh.cells("points"), [[0, 3, 2], [0, 2, 1]])
    assert np.array_equal(mesh.cells("edges"), [[4, 3, 1], [2, 0, 3]])


def test_flip_simple_opposite_orientation():
    #        3                   3
    #        A                   A
    #       /|\                 / \
    #     1/ | \4             1/ 1 \4
    #     /  |  \             /     \
    #   0/ 0 3   \2   ==>   0/___3___\2
    #    \   | 1 /           \       /
    #     \  |  /             \     /
    #     0\ | /2             0\ 0 /2
    #       \|/                 \ /
    #        V                   V
    #        1                   1
    #
    points = np.array([[-0.1, 0.0], [0.0, -1.0], [0.1, 0.0], [0.0, 1.1]])
    cells = np.array([[0, 1, 3], [1, 3, 2]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.create_facets()
    assert mesh.num_delaunay_violations == 1
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [1, 3], [2, 3]]
    )
    assert np.array_equal(mesh.cells("edges"), [[3, 1, 0], [4, 2, 3]])
    assert_mesh_consistency(mesh)

    # mesh.show()
    mesh.flip_until_delaunay()
    assert_mesh_consistency(mesh)
    assert mesh.num_delaunay_violations == 0
    assert np.array_equal(
        mesh.edges["points"], [[0, 1], [0, 3], [1, 2], [0, 2], [2, 3]]
    )
    assert np.array_equal(mesh.cells("points"), [[0, 1, 2], [0, 2, 3]])
    assert np.array_equal(mesh.cells("edges"), [[2, 3, 0], [4, 1, 3]])


def test_flip_delaunay_near_boundary():
    points = np.array([[0.0, +0.0], [0.5, -0.1], [1.0, +0.0], [0.5, +0.1]])
    cells = np.array([[0, 1, 2], [0, 2, 3]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.create_facets()
    assert mesh.num_delaunay_violations == 1
    assert np.array_equal(mesh.cells("points"), [[0, 1, 2], [0, 2, 3]])
    assert np.array_equal(mesh.cells("edges"), [[3, 1, 0], [4, 2, 1]])

    mesh.flip_until_delaunay()

    assert_mesh_consistency(mesh)
    assert mesh.num_delaunay_violations == 0
    assert np.array_equal(mesh.cells("points"), [[1, 2, 3], [1, 3, 0]])
    assert np.array_equal(mesh.cells("edges"), [[4, 1, 3], [2, 0, 1]])


def test_flip_same_edge_twice():
    points = np.array([[0.0, +0.0], [0.5, -0.1], [1.0, +0.0], [0.5, +0.1]])
    cells = np.array([[0, 1, 2], [0, 2, 3]])
    mesh = meshplex.MeshTri(points, cells)
    assert mesh.num_delaunay_violations == 1

    mesh.flip_until_delaunay()
    assert mesh.num_delaunay_violations == 0

    mesh.show(
        mark_cells=mesh.is_boundary_cell,
        show_point_numbers=True,
        show_edge_numbers=True,
        show_cell_numbers=True,
    )
    assert_mesh_consistency(mesh)

    new_points = np.array([[0.0, +0.0], [0.1, -0.5], [0.2, +0.0], [0.1, +0.5]])
    mesh.points = new_points
    assert mesh.num_delaunay_violations == 1

    mesh.flip_until_delaunay()
    assert mesh.num_delaunay_violations == 0
    mesh.show()
    # mesh.plot()


def test_flip_two_edges():
    alpha = np.array([1.0, 3.0, 5.0, 7.0, 9.0, 11.0]) / 6.0 * np.pi
    # Make the mesh slightly asymmetric to get the same flips on every architecture; see
    # <https://github.com/nschloe/meshplex/issues/78>.
    R = [0.95, 1.0, 0.9, 1.0, 1.2, 1.0]
    points = np.array([[r * np.cos(a), r * np.sin(a), 0.0] for a, r in zip(alpha, R)])
    cells = np.array([[1, 3, 5], [0, 1, 5], [1, 2, 3], [3, 4, 5]])
    mesh = meshplex.MeshTri(points, cells)
    assert mesh.num_delaunay_violations == 2

    mesh.flip_until_delaunay()
    assert mesh.num_delaunay_violations == 0

    mesh.show(show_point_numbers=True)
    assert np.array_equal(
        mesh.cells("points"), [[2, 5, 0], [2, 0, 1], [5, 2, 3], [3, 4, 5]]
    )


def test_flip_delaunay_near_boundary_preserve_boundary_count():
    # This test is to make sure meshplex preserves the boundary point count.
    points = np.array(
        [
            [+0.0, +0.0],
            [+0.5, -0.5],
            [+0.5, +0.5],
            [+0.0, +0.6],
            [-0.5, +0.5],
            [-0.5, -0.5],
        ]
    )
    cells = np.array([[0, 1, 2], [0, 2, 4], [0, 4, 5], [0, 5, 1], [2, 3, 4]])
    mesh = meshplex.MeshTri(points, cells)

    mesh.create_facets()
    assert mesh.num_delaunay_violations == 1

    is_boundary_point_ref = [False, True, True, True, True, True]
    assert np.array_equal(mesh.is_boundary_point, is_boundary_point_ref)

    mesh.flip_until_delaunay()
    assert np.array_equal(mesh.is_boundary_point, is_boundary_point_ref)


def test_flip_orientation():
    points = np.array([[0.0, +0.0], [0.5, -0.1], [1.0, +0.0], [0.5, +0.1]])

    # preserve positive orientation
    cells = np.array([[0, 1, 2], [0, 2, 3]])
    mesh = meshplex.MeshTri(points, cells)
    assert np.all(mesh.signed_cell_volumes > 0.0)
    mesh.flip_until_delaunay()
    assert np.all(mesh.signed_cell_volumes > 0.0)

    # also preserve negative orientation
    cells = np.array([[0, 2, 1], [0, 3, 2]])
    mesh = meshplex.MeshTri(points, cells)
    assert np.all(mesh.signed_cell_volumes < 0.0)
    mesh.flip_until_delaunay()
    assert np.all(mesh.signed_cell_volumes < 0.0)


def test_flip_infinite():
    """In rare cases, it can happen that the ce-ratio of an edge is negative (up to
    machine precision, -2.13e-15 or something like that), an edge flip is done, and the
    ce-ratio of the resulting edge is again negative. The flip_until_delaunay() method
    would continue indefinitely. This test replicates such an edge case."""
    a = 3.9375644347017862e02
    points = np.array([[205.0, a], [185.0, a], [330.0, 380.0], [60.0, 380.0]])
    cells = [[0, 1, 2], [1, 2, 3]]

    mesh = meshplex.MeshTri(points, cells)
    num_flips = mesh.flip_until_delaunay(tol=1.0e-13)
    assert num_flips == 0


def test_flip_interior_to_boundary():
    #  __________          __________
    #  |\__      A         |\__      A
    #  |   \__  /|\        |   \__  / \
    #  |      \/ | \  ==>  |      \/___\
    #  |    __/\ | /       |    __/\   /
    #  | __/    \|/        | __/    \ /
    #  |/________V         |/________V
    #
    points = np.array(
        [[0.0, 0.0], [1.0, 0.0], [1.1, 0.5], [1.0, 1.0], [0.0, 1.0], [0.9, 0.5]]
    )
    cells = np.array([[0, 1, 5], [1, 3, 5], [1, 2, 3], [3, 4, 5], [0, 5, 4]])

    mesh = meshplex.MeshTri(points, cells)
    compute_all_entities(mesh)
    # mesh.show(mark_cells=mesh.is_boundary_cell)
    mesh.flip_until_delaunay()
    assert_mesh_consistency(mesh)
    # mesh.show(mark_cells=mesh.is_boundary_cell)
    assert np.all(mesh.is_boundary_cell)


def test_flip_delaunay():
    rng = np.random.default_rng(123)
    mesh0 = meshio.read(this_dir / ".." / "meshes" / "pacman.vtu")
    mesh0.points[:, :2] += 1.0e-1 * rng.random(mesh0.points[:, :2].shape)

    mesh0 = meshplex.MeshTri(mesh0.points[:, :2], mesh0.get_cells_type("triangle"))
    compute_all_entities(mesh0)

    assert np.all(mesh0.signed_cell_volumes > 0)

    assert mesh0.num_delaunay_violations == 5

    mesh0.flip_until_delaunay()
    assert mesh0.num_delaunay_violations == 0

    assert_mesh_consistency(mesh0)

    # mesh0.show(mark_cells=mesh0.is_boundary_cell)

    # We don't need to check for exact equality with a replicated mesh. The order of the
    # edges will be different, for example. Just make sure the mesh is consistent.
    # mesh1 = meshplex.MeshTri(mesh0.points.copy(), mesh0.cells("points").copy())
    # mesh1.create_facets()
    # assert_mesh_equality(mesh0, mesh1)


def test_flip_into_existing_edge():
    """For surface meshes, flips can lead to duplicate cells. For context, see
    <https://github.com/nschloe/optimesh/issues/71#issuecomment-785699560>.
    """
    points = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.5, 0.3, 0.3],
            [0.5, -0.3, 0.3],
        ]
    )
    cells = np.array(
        [
            [1, 2, 0],
            [2, 3, 0],
            [3, 1, 0],
        ]
    )

    mesh = meshplex.MeshTri(points, cells)
    with pytest.warns(UserWarning):
        n = mesh.flip_until_delaunay()
    # no flips performed
    assert n == 0


def test_doubled_cell():
    # Two congruent cells. One can think of it as a deflated, coarse ball.
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.5, 0.4],
        ]
    )
    cells = np.array([[0, 1, 2], [0, 1, 2]])
    mesh = meshplex.MeshTri(points, cells)

    num_flips = mesh.flip_until_delaunay()
    assert num_flips == 1

    ref = [[2, 0, 2], [2, 2, 1]]
    assert np.all(mesh.cells("points") == ref)


def test_negative_after_flip():
    points = [[0.0, 0.0], [3.0, 0.0], [1.14960653, 0.03], [1.85039347, 0.03]]
    cells = [
        [0, 3, 2],
        [0, 1, 3],
    ]
    mesh0 = meshplex.MeshTri(points, cells)

    with pytest.warns(UserWarning):
        mesh0.flip_until_delaunay()



import pytest

import meshplex


def test_gh126():
    """https://github.com/nschloe/meshplex/issues/126"""
    cells = [[1, 2, 0], [2, 3, 0], [3, 1, 0]]
    points = [
        [4.3000283, 8.57424769, 0.010431],
        [4.2568793, 8.41702649, 0.0],
        [4.2864766, 8.56681008, 0.0],
        [4.2897457, 8.58639357, 0.0],
    ]
    mesh = meshplex.MeshTri(points, cells)

    with pytest.warns(UserWarning):
        mesh.flip_until_delaunay()

    # this failed in a previous version because there were edges with more than two
    # cells after the flip
    mesh.create_facets()


def test_gh130():
    """https://github.com/nschloe/meshplex/issues/130"""
    points = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.0],
        [1.0, 0.9, 1.0],
        [1.0, 1.1, 1.0],
        [1.0, 0.7, 1.0],
        [1.0, 0.7, 0.0],
    ]
    cells = [[2, 0, 1], [2, 3, 0], [4, 0, 3], [4, 3, 5], [2, 5, 3]]

    mesh = meshplex.MeshTri(points, cells)
    with pytest.warns(UserWarning):
        mesh.flip_until_delaunay(max_steps=1)

    # this failed in a previous version because there were edges with more than two
    # cells after the flip
    mesh.create_facets()



import pathlib

import meshplex

this_dir = pathlib.Path(__file__).resolve().parent


def test_get_edges():
    mesh = meshplex.read(this_dir / "meshes" / "pacman.vtu")
    mesh.create_facets()
    edge_mask = mesh.get_edge_mask()
    edge_points = mesh.edges["points"][edge_mask]
    assert len(edge_points) == 2372


def test_mark_subdomain2d():
    mesh = meshplex.read(this_dir / "meshes" / "pacman.vtu")

    class Subdomain1:
        is_boundary_only = True

        # pylint: disable=no-self-use
        def is_inside(self, x):
            return x[0] < 0.0

    class Subdomain2:
        is_boundary_only = False

        # pylint: disable=no-self-use
        def is_inside(self, x):
            return x[0] > 0.0

    sd1 = Subdomain1()
    vertex_mask = mesh.get_vertex_mask(sd1)
    assert vertex_mask.sum() == 45
    face_mask = mesh.get_face_mask(sd1)
    assert face_mask.sum() == 44
    cell_mask = mesh.get_cell_mask(sd1)
    assert cell_mask.sum() == 0

    sd2 = Subdomain2()
    vertex_mask = mesh.get_vertex_mask(sd2)
    assert vertex_mask.sum() == 395
    face_mask = mesh.get_face_mask(sd2)
    assert face_mask.sum() == 2148
    cell_mask = mesh.get_cell_mask(sd2)
    assert cell_mask.sum() == 706


def test_mark_subdomain3d():
    mesh = meshplex.read(this_dir / "meshes" / "tetrahedron.vtk")

    class Subdomain1:
        is_boundary_only = True

        # pylint: disable=no-self-use
        def is_inside(self, x):
            return x[0] < 0.5

    class Subdomain2:
        is_boundary_only = False

        # pylint: disable=no-self-use
        def is_inside(self, x):
            return x[0] > 0.5

    sd1 = Subdomain1()
    vertex_mask = mesh.get_vertex_mask(sd1)
    assert vertex_mask.sum() == 16
    face_mask = mesh.get_face_mask(sd1)
    assert face_mask.sum() == 20
    cell_mask = mesh.get_cell_mask(sd1)
    assert cell_mask.sum() == 0

    sd2 = Subdomain2()
    vertex_mask = mesh.get_vertex_mask(sd2)
    assert vertex_mask.sum() == 10
    face_mask = mesh.get_face_mask(sd2)
    assert face_mask.sum() == 25
    cell_mask = mesh.get_cell_mask(sd2)
    assert cell_mask.sum() == 5



from math import fsum

import numpy as np


def is_near_equal(a, b, tol):
    return np.allclose(a, b, rtol=0.0, atol=tol)


def run(mesh, volume, convol_norms, ce_ratio_norms, cellvol_norms, tol=1.0e-12):
    # Check cell volumes.
    total_cellvolume = fsum(mesh.cell_volumes)
    assert abs(volume - total_cellvolume) < tol * volume, total_cellvolume
    norm2 = np.linalg.norm(mesh.cell_volumes, ord=2)
    norm_inf = np.linalg.norm(mesh.cell_volumes, ord=np.Inf)
    assert is_near_equal(cellvol_norms, [norm2, norm_inf], tol), [norm2, norm_inf]

    # If everything is Delaunay and the boundary elements aren't flat, the volume of the
    # domain is given by
    #   1/n * edge_lengths * ce_ratios.
    # Unfortunately, this isn't always the case.
    # ```
    # total_ce_ratio = \
    #     fsum(mesh.edge_lengths**2 * mesh.get_ce_ratios_per_edge() / dim)
    # self.assertAlmostEqual(volume, total_ce_ratio, delta=tol * volume)
    # ```
    # Check ce_ratio norms.
    alpha2 = fsum((mesh.ce_ratios**2).flat)
    alpha_inf = max(abs(mesh.ce_ratios).flat)
    assert is_near_equal(ce_ratio_norms, [alpha2, alpha_inf], tol), [alpha2, alpha_inf]

    # Check the volume by summing over the absolute value of the control volumes.
    vol = fsum(mesh.control_volumes)
    assert abs(volume - vol) < tol * volume, vol
    # Check control volume norms.
    norm2 = np.linalg.norm(mesh.control_volumes, ord=2)
    norm_inf = np.linalg.norm(mesh.control_volumes, ord=np.Inf)
    assert is_near_equal(convol_norms, [norm2, norm_inf], tol), [norm2, norm_inf]


def assert_norms(x, ref, tol):
    ref = np.asarray(ref)
    val = np.array(
        [
            np.linalg.norm(x.flat, 1),
            np.linalg.norm(x.flat, 2),
            np.linalg.norm(x.flat, np.inf),
        ]
    )
    assert np.all(np.abs(val - ref) < tol * np.abs(ref)), (
        "Norms don't coincide.\n"
        f"Expected: [{ref[0]:.16e}, {ref[1]:.16e}, {ref[2]:.16e}]\n"
        f"Computed: [{val[0]:.16e}, {val[1]:.16e}, {val[2]:.16e}]\n"
    )



import numpy as np

import meshplex


def _is_near_equal(a, b, tol=1.0e-12):
    return np.allclose(a, b, rtol=0.0, atol=tol)


def test_mesh_line():
    pts = [0.0, 1.0, 3.0, 4.0]
    cells = [[0, 1], [1, 2], [2, 3]]
    mesh = meshplex.Mesh(pts, cells)
    assert _is_near_equal(mesh.cell_volumes, [1.0, 2.0, 1.0])
    assert _is_near_equal(mesh.control_volumes, [0.5, 1.5, 1.5, 0.5])

    assert np.all(mesh.is_boundary_point == [True, False, False, True])

    # control volume with some index
    c = mesh.get_control_volume_centroids([False, False, False])
    assert _is_near_equal(c, [0.25, 1.25, 2.75, 3.75])



import pathlib

import numpy as np

import meshplex

this_dir = pathlib.Path(__file__).resolve().parent


def test_pacman():
    mesh = meshplex.read(this_dir / ".." / "meshes" / "pacman.vtu")

    # mesh = meshplex.MeshTri(mesh.points[:, :2], mesh.cells("points"))
    # mesh.signed_cell_volumes
    # exit(1)

    # Create circular vector field 0.5 * (y, -x, 0)
    # which has curl (0, 0, 1).
    A = np.array([[-0.5 * coord[1], 0.5 * coord[0], 0.0] for coord in mesh.points])
    # Compute the curl numerically.
    B = mesh.compute_ncurl(A)

    # mesh.write(
    #     'curl.vtu',
    #     point_data={'A': A},
    #     cell_data={'B': B}
    #     )

    tol = 1.0e-14
    for b in B:
        assert abs(b[0] - 0.0) < tol
        assert abs(b[1] - 0.0) < tol
        assert abs(b[2] - 1.0) < tol


if __name__ == "__main__":
    test_pacman()
